# Copyright 2023 Lawrence Livermore National Security, LLC
# (c.f. AUTHORS, NOTICE.LLNS, COPYING)

from .network import network_netmark, network_osu_benchmark
from .storage import io_sysstat

metrics = {
    "io-sysstat": io_sysstat,
    "network-osu-benchmark": network_osu_benchmark,
    "network-netmark": network_netmark,
}


def get_metric(name):
    """
    Get a named metric parser.
    """
    metric = metrics.get(name)
    if not metric:
        raise ValueError(f"Metric {name} does not have a known parser")
    return metric



# Copyright 2023 Lawrence Livermore National Security, LLC
# (c.f. AUTHORS, NOTICE.LLNS, COPYING)

import json

from .base import MetricBase


class io_sysstat(MetricBase):
    container_name = "io-sysstat"

    @property
    def pod_prefix(self):
        return f"{self.name}-m-0"

    def parse_log(self, lines):
        """
        Given lines of output, parse and return json
        """
        # Get the log metadata
        metadata = self.get_log_metadata(lines)

        # Split lines by IOSTAT TIMEPOINT
        results = []
        sections = self.get_log_sections(lines)
        for section in sections:
            if not section.strip():
                continue
            results.append(json.loads(section))
        return {"data": results, "metadata": metadata, "spec": self.spec}



# Copyright 2023 Lawrence Livermore National Security, LLC
# (c.f. AUTHORS, NOTICE.LLNS, COPYING)


from .base import MetricBase


class network_osu_benchmark(MetricBase):
    """
    Parse the OSU benchmarks output into data!
    """

    container_name = "launcher"

    @property
    def pod_prefix(self):
        return f"{self.name}-l-0"

    def parse_row(self, row):
        """
        Given a row of two values with spaces, parse.
        """
        row = row.split(" ", 1)
        return [x.strip() for x in row]

    def parse_log(self, lines):
        """
        Given lines of output, parse and return json
        """
        # Get the log metadata
        metadata = self.get_log_metadata(lines)

        # Split lines by separator
        results = []
        sections = self.get_log_sections(lines)
        for section in sections:
            if not section.strip():
                continue
            section = section.split("\n")
            section = [x.strip() for x in section if x.strip()]

            # Command is the first entry
            command = section.pop(0)

            # Each section has some number of header lines (with #)
            header = []
            while section[0].startswith("#"):
                header.append(section.pop(0))

            # Last row of the header are the column names
            columns = header.pop()
            columns = columns.replace("# ", "").strip()
            columns = self.parse_row(columns)

            # The remainder is data, again always two points
            data = []
            for line in section:
                if not line:
                    continue
                row = self.parse_row(line)
                row = [float(x) for x in row]
                data.append(row)

            datum = {
                "matrix": data,
                "columns": columns,
                "header": header,
                "command": command,
            }
            results.append(datum)

        return {"data": results, "metadata": metadata, "spec": self.spec}


class network_netmark(MetricBase):
    """
    Parse the netmark terminal output.
    """

    container_name = "launcher"

    @property
    def pod_prefix(self):
        return f"{self.name}-n-0"

    def parse_row(self, row):
        """
        Given a row of netmark output, parse!
        """
        row = row.split(" ", 1)
        return [x.strip() for x in row]

    def parse_log(self, lines):
        """
        Given lines of output, parse and return json
        """
        # Get the log metadata
        metadata = self.get_log_metadata(lines)

        # Split lines by separator
        results = []
        sections = self.get_log_sections(lines)

        # Netmark just has one long run (one section)
        for section in sections:
            section = section.split("\n")
            section = [x.strip() for x in section if x.strip()]

            # The first lines up to SETUP have information about ranks
            ranks = []
            while "SETUP" not in section[0]:
                ranks.append(section.pop(0))

            # Next is setup section (we have this in our metadata)
            section.pop(0)
            setup = []
            while "======" not in section[0]:
                setup.append(section.pop(0))

            # Pop end of setup line
            section.pop(0)

            # Find the RTT.csv between netmark.start and end lines
            section = "\n".join(section)
            netmark_data = section.split("NETMARK RTT.CSV START", 1)[-1]
            netmark_data = netmark_data.split("NETMARK RTT.CSV END", 1)[0]

            # Add rest of ranks from original data
            section = section.split("\n")
            while section:
                if section[0].startswith("size"):
                    ranks.append(section.pop(0))
                else:
                    section.pop(0)

            datum = {
                "RTT.csv": netmark_data.strip(),
                "ranks": ranks,
                "setup": setup,
            }
            results.append(datum)

        return {"data": results, "metadata": metadata, "spec": self.spec}



import json
import time

from kubernetes import client, config, watch
from kubernetes.client.api import core_v1_api
from kubernetes.client.exceptions import ApiException
from kubernetes.client.models.v1_pod_list import V1PodList


class MetricBase:
    separator = "METRICS OPERATOR TIMEPOINT"
    collection_start = "METRICS OPERATOR COLLECTION START"
    collection_end = "METRICS OPERATOR COLLECTION END"
    metadata_start = "METADATA START"
    metadata_end = "METADATA END"

    def __init__(self, spec=None, **kwargs):
        """
        Create a persistent client to interact with a MiniCluster

        This currently assumes the namespace exists.
        """
        self.spec = spec
        self._core_v1 = kwargs.get("core_v1_api")

        # Load kubeconfig on Metricbase init only
        if self.spec is not None:
            config.load_kube_config()

    @property
    def namespace(self):
        if not self.spec:
            return
        return self.spec["metadata"].get("namespace") or "default"

    @property
    def name(self):
        if not self.spec:
            return
        return self.spec["metadata"]["name"]

    @property
    def classname(self):
        return self.__class__.__name__

    def container(self):
        return self.classname.replace("-", "_")

    def parse(self, pod, container):
        """
        Retrieve logs output and call parsing function
        """
        lines = self.stream_output(
            name=pod.metadata.name, namespace=self.namespace, container=container.name
        )
        return self.parse_log(lines)

    @property
    def core_v1(self):
        """
        Instantiate a core_v1 api (if not done yet)
        """
        if self._core_v1 is not None:
            return self._core_v1

        self.c = client.Configuration.get_default_copy()
        self.c.assert_hostname = False
        client.Configuration.set_default(self.c)
        self._core_v1 = core_v1_api.CoreV1Api()
        return self._core_v1

    def logging_containers(self, namespace=None, states=None, retry_seconds=5):
        """
        Return list of containers intended to get logs from
        """
        containers = []
        pods = self.wait(namespace, states, retry_seconds)
        container_name = getattr(self, "container_name", self.container)
        print(f"Looking for container name {container_name}...")
        for pod in pods.items:
            for container in pod.spec.containers:
                print(f"Assessing {container.name}")
                if container.name == container_name:
                    print(f"Found logging container {container.name}")
                    containers.append(
                        (
                            pod,
                            container,
                        )
                    )
        return containers

    def wait(self, namespace=None, states=None, retry_seconds=5):
        """
        Wait for one or more pods of interest to be done
        """
        namespace = namespace or self.namespace
        print(f"Looking for prefix {self.pod_prefix} in namespace {namespace}")
        pod_list = self.get_pods(namespace, self.pod_prefix)
        size = len(pod_list.items)

        # We only want logs when they are completed
        states = states or ["Completed", "Succeeded"]
        if not isinstance(states, list):
            states = [states]

        ready = set()
        while len(ready) != size:
            print(f"{len(ready)} pods are ready, out of {size}")
            pod_list = self.get_pods(name=self.pod_prefix, namespace=namespace)

            for pod in pod_list.items:
                print(f"{pod.metadata.name} is in phase {pod.status.phase}")
                if pod.status.phase not in states:
                    time.sleep(retry_seconds)
                    continue

                if pod.status.phase not in ["Terminating"]:
                    ready.add(pod.metadata.name)

        states = '" or "'.join(states)
        print(f'All pods are in states "{states}"')
        return pod_list

    def _filter_pods(self, pods, name):
        """
        Filter a set of pods (associated with a job) to a name prefix.
        """
        filtered = []
        for pod in pods.items:
            if pod.metadata.name.startswith(name):
                filtered.append(pod)
        pods.items = filtered
        return pods

    def get_pods(self, namespace=None, name=None):
        """
        Get namespaced pods metadata, either scoped to a name or entire namespace.
        """
        namespace = namespace or self.namespace
        try:
            req = self.core_v1.list_namespaced_pod(namespace, async_req=True)
            pods = req.get()

            # If name is present, filter down to pods with that prefix
            if name is not None:
                pods = self._filter_pods(pods, name)
            return pods

        # Not found - it was deleted
        except ApiException:
            return V1PodList(items=[])
        except Exception:
            time.sleep(2)
            return self.get_pods(namespace, name)

    def get_log_metadata(self, lines):
        """
        Given a log dump, split based on the known separators.
        """
        if self.metadata_start not in lines or self.metadata_end not in lines:
            print("Cannot find expected collection start or end lines, cannot parse")
            return {}
        metadata = lines.split(self.metadata_start, 1)[-1]
        metadata = metadata.split(self.metadata_end, 1)[0]
        return json.loads(metadata)

    def get_log_sections(self, lines):
        """
        Given a log dump, split into data sections
        """
        if self.collection_start not in lines or self.collection_end not in lines:
            print("Cannot find expected metadata start or end lines, cannot parse")
            return {}
        data = lines.split(self.collection_start, 1)[1:]
        data = "\n".join(data).split(self.collection_end, 1)[0]
        return data.split(self.separator)

    def stream_output(
        self,
        name,
        namespace,
        stdout=True,
        filename=None,
        container=None,
        timestamps=False,
    ):
        """
        Stream output, optionally printing also to stdout.

        Also return the output to the user.
        """
        watcher = watch.Watch()

        out = None
        if filename:
            out = open(filename, "w")

        # Stream output to file and return it if desired!
        lines = []
        for line in watcher.stream(
            self.core_v1.read_namespaced_pod_log,
            name=name,
            namespace=namespace,
            timestamps=timestamps,
            container=container,
            follow=True,
        ):
            if out:
                # Lines end with /r and we need to strip and add a newline
                out.write(line.strip() + "\n")
            if stdout:
                print(line)
            lines.append(line)

        if out:
            out.close()

        # The parser needs to split, so we expect a single cohesive string
        return "\n".join(lines)



# Copyright 2023 Lawrence Livermore National Security, LLC
# (c.f. AUTHORS, NOTICE.LLNS, COPYING)

import json

import yaml


def read_yaml(filename):
    """
    Read yaml into dict
    """
    with open(filename, "r") as file:
        configuration = yaml.safe_load(file)
    return configuration


def write_json(obj, filename):
    """
    Write dict to json
    """
    with open(filename, "w") as fd:
        fd.write(json.dumps(obj, indent=4))


def read_json(filename):
    """
    Read json into dict
    """
    with open(filename, "r") as fd:
        content = json.loads(fd.read())
    return content
